"""
Classes and functions used for the communication between Z3Compiler and users
"""
import typing
from typing import NamedTuple, Union
from enum import Enum, unique


class Explanation(NamedTuple):
    text: str = "We could not generate an explanation to describe why the execution is invalid. This is a shortcoming of the tool, please report it!"
    challengable: str = ""
    substitutions: list = []


@unique
class MemoryModel(Enum):
    """Enum to select which memory model to use"""
    SequentialConsistency = "seqConsist"
    WeakConsistency = "weak"
    WeakConsistencyFences = "weak_fences"
    WeakConsistencyLocks = "weak_locks"


class InstructionId(NamedTuple):
    agent_id: int
    statement_id: int

    @staticmethod
    def from_id_string(id_str: str):
        assert id_str.startswith("a")
        return InstructionId(*id_str[1:].split(":"))


class OrderAssertion(NamedTuple):
    """Describes asserttion InstructionOrder(predecessor, statement)
    Implicitly marks statement and predecessor as Reachable
    """
    id: int
    successor: InstructionId
    predecessor: InstructionId


class WriteSeenAssertion(NamedTuple):
    """Describes asserttion WriteSeen(writer, reader)
    Implicitly marks writer and reader as Reachable
    """
    id: int
    writer: InstructionId
    reader: InstructionId
    variable: str


class ReadValueAssertion(NamedTuple):
    """Describes assertion ReadValue(instruction, variable, value)
    Implicitly marks instruction as Reachable
    """
    id: int
    instruction: InstructionId
    variable: str
    value: int


class WriteValueAssertion(NamedTuple):
    """Describes assertion WriteValue(instruction, variable, value)
    Implicitly marks instruction as Reachable
    """
    id: int
    instruction: InstructionId
    variable: str
    value: int


Assertion = Union[OrderAssertion, WriteSeenAssertion, ReadValueAssertion, WriteValueAssertion]
