"""The m3 server, which handles all requests over WebSockets and delegates the actual work to the
pseucparser and to the Z3Compiler"""

import ssl
import itertools
import functools
import time
import json
import asyncio
import multiprocessing
import logging
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from pyparsing import ParseException

from autobahn.asyncio.websocket import WebSocketServerProtocol, \
    WebSocketServerFactory
from autobahn.websocket.compress import PerMessageDeflateOffer, \
    PerMessageDeflateOfferAccept

from pseuco_parser import createParser, UnsupportedPseuCoException
from z3compiler import Z3Compiler, CreateNewModel
from communication import (Explanation, MemoryModel, InstructionId, OrderAssertion,
                           WriteSeenAssertion, ReadValueAssertion, WriteValueAssertion)


def get_cfg_for_program(program):
    """Creates a cfg representation of the given program"""
    ret = {}
    counter = 0
    for agent_name in itertools.chain.from_iterable(
            program.startedAgentNames.asList()):
        agent = program.initialization.agent_declarations[agent_name]
        body = agent.functionBody
        cfg_predecessors = []
        cfc_graph_representation = []
        ret["init:0"] = program.initialization.create_node_data()
        for statement in body.statements:
            cfc_graph_representation += statement.create_node_data(
                cfg_predecessors)
            cfg_predecessors = statement.get_cfg_exit_nodes()
        ret[agent_name + ":" + str(counter+1)] = cfc_graph_representation
        counter += 1
    return ret


class BookKeeper:
    """Keeps track of how many ressources a client is using,
       Currently broken, and thus not used
    """
    def __init__(self, score_limit=1000):
        """Sets up the bookkeeping"""
        self.ip2score = defaultdict(int)
        self.ip2last_access = {}
        self.time_of_last_reset = time.monotonic()
        self.limit = score_limit

    def _improve_score(self):
        time_since_reset = time.monotonic() - self.time_of_last_reset
        score_reduction = int(time_since_reset) / (60000 * 30)  # 15 minutes
        if time_since_reset <= 0:
            return
        self.time_of_last_reset = time.monotonic()
        improvement = int(time_since_reset)
        removable_users = []
        for ip, score in self.ip2score.items():
            new_score = max(0, score - 100*score_reduction)
            if new_score == 0:
                removable_users.append(ip)
            else:
                self.ip2score[ip] = new_score
        for user in removable_users:
            del self.ip2last_access[user]
            del self.ip2score[user]

    def caused_timeout(self, ip):
        """Called when a user caused a timeout of the z3 solver"""

    def track_and_check_visit(self, ip):
        """Call when user visits site"""
        self._improve_score()
        self.ip2last_access[ip] = time.monotonic()
        return self.ip2score[ip] > self.limit

    def update_score(self, ip):
        last_access = time.monotonic()
        time_spent = int(self.ip2last_access[ip] - last_access)
        self.ip2score[ip] += (time_spent+1)*50

book_keeper = BookKeeper()


class M3ServerProtocoll(WebSocketServerProtocol):

    def _ack_nack_helper(func):
        """Easier message handling"""
        @functools.wraps(func)
        async def wrapper(self, *args, **kwargs):
            success, id, explanation = await func(self, *args, **kwargs)
            logging.info(f"{func.__name__}, {success}, {id}, {explanation}")
            if success:
                return {
                    "type": "ack",
                    "id": id
                }
            else:
                return {
                    "type": "nack",
                    "id": id,
                    "explanation": explanation
                }

        return wrapper


    executor = ThreadPoolExecutor(max_workers=multiprocessing.cpu_count(),
                                  thread_name_prefix="z3_task")

    mem_model = {
        "seqConsist": MemoryModel.SequentialConsistency,
        "weak": MemoryModel.WeakConsistency
    }

    def __init__(self):
        super().__init__()
        self.handle = {
            "code": self._onCodeMessage,
            "WriteSeenAdded": self._onWriteSeenAdded,
            "setMemoryModel": self._onsetMemoryModel,
            "switchAssertionOrder": self._onSwitchAssertionOrder,
            "valueSet": self._onValueSet,
            "writeValueSet": self._onWriteValueSet,
            "OrderAdded": self._onOrderAdded,
            "undoAssertion": self._onUndoAssertion,
        }
        self.outstanding_work = None
        self.compiler = None
        self.executions = None
        self.seq_num = 0

    async def _onUndoAssertion(self, decoded_message):
        id_ = decoded_message["id"]
        self.compiler.remove_assertion(id_)
        reply = {
            "type": "removedAssertion",
            "id": id_
        }
        return reply

    async def _onCodeMessage(self, decoded_message):
        try:
            logging.info("got message")
            tinyPseucoParser = createParser()
            program = tinyPseucoParser.parseString(decoded_message["code"], parseAll=True)
            if type(program.startedAgentNames) is str:
                raise RuntimeWarning("You must start at least one agent, else analysing"
                                                 " the memory model does not make any sense!")
            reply = {
                "cfg": get_cfg_for_program(program),
                "type": "parseResult"
            }
            loop = asyncio.get_event_loop()
            self.outstanding_work = loop.run_in_executor(
                M3ServerProtocoll.executor,
                Z3Compiler,
                program,
            )
        except (UnsupportedPseuCoException, ParseException, RuntimeWarning, RuntimeError) as e:
            logging.info("Parsing failed: " + e.msg)
            if "end of text" in e.msg:
                reply = {
                    "type": "error",
                    "errorMessage": str(e) + "\n" + "(Ensure that the mainAgent is called after all"
                                                    " agents are declared)"
                }
            else:
                reply = {
                    "type": "error",
                    "errorMessage": str(e),
                }
        return reply

    async def _onGetAllExecutions(self, decoded_message):
        # loop = asyncio.get_event_loop()
        # allReads = await loop.run_in_executor(
        #     M3ServerProtocoll.executor,
        #     self.compiler.list_executions,
        # )
        strategy = CreateNewModel(0)
        if "strategy" in decoded_message and  decoded_message["strategy"] != "ignoreNothing":
            logging.info("Considering all order permutations")
            strategy |= CreateNewModel.IgnoreOrder
        if "strategy" in decoded_message and  decoded_message["strategy"] == "ignoreWriteSeenAndOrder":
            logging.info("Discarding executions which only differ in their write seen relationship")
            strategy |= CreateNewModel.IgnoreWriteSeenRel
        for execution in self.compiler.list_executions(strategy):
            yield {
                "type": "newExecution",
                "execution": execution,
            }
        yield {
            "type": "allExecutionsListed",
        }

    async def _onsetMemoryModel(self, decoded_message):
        mem_model = MemoryModel(decoded_message["memoryModel"])
        acked, nacked = self.compiler.set_memory_model(mem_model)
        return {
            "type": "memoryModelChanged",
            "acked": acked,
            "nacked": nacked,
        }

    async def _onSwitchAssertionOrder(self, decoded_message):
        oldpos = int(decoded_message["oldPos"])
        newpos = int(decoded_message["newPos"])
        acked, nacked = self.compiler.switch_manually_added_assertion(oldpos, newpos)
        # We can reuse memoryModelChanged
        return {
            "type": "memoryModelChanged",
            "acked": acked,
            "nacked": nacked,
        }

    @_ack_nack_helper
    async def _onWriteSeenAdded(self, decoded_message):
        read = InstructionId(decoded_message["seen"]["agent"],
                             decoded_message["seen"]["statement"])
        write = InstructionId(decoded_message["write"]["agent"],
                              decoded_message["write"]["statement"])
        write_seen = WriteSeenAssertion(writer=write,
                                        id=decoded_message["id"],
                                        reader=read,
                                        variable=decoded_message["varname"])
        sat, explanation = self.compiler.add_constraint(write_seen)
        return sat, decoded_message["id"], explanation

    @_ack_nack_helper
    async def _onOrderAdded(self, decoded_message):
        prev = InstructionId(decoded_message["prev"]["agent"],
                            decoded_message["prev"]["statement"])
        succ = InstructionId(decoded_message["next"]["agent"],
                            decoded_message["next"]["statement"])
        # ids start at 0 for initialization, but its 1 instruction2id
        if prev == InstructionId(0,0):
            prev = InstructionId(0,1)
        if succ == InstructionId(0,0):
            succ = InstructionId(0,1)
        order = OrderAssertion(successor=succ, predecessor=prev, id=decoded_message["id"])
        sat, explanation = self.compiler.add_constraint(order)
        return sat, decoded_message["id"], explanation

    @_ack_nack_helper
    async def _onValueSet(self, decoded_message):
        value = int(decoded_message["value"])
        variable = decoded_message["varname"]
        instruction = InstructionId(decoded_message["statement"]["agent"],
                                    decoded_message["statement"]["statement"]
                                   )
        read_value = ReadValueAssertion(instruction=instruction,
                                        variable=variable,
                                        value=value,
                                        id=decoded_message["id"])
        sat, explanation = self.compiler.add_constraint(read_value)
        return sat, decoded_message["id"], explanation

    @_ack_nack_helper
    async def _onWriteValueSet(self, decoded_message):
        value = int(decoded_message["value"])
        variable = decoded_message["varname"]
        instruction = InstructionId(decoded_message["statement"]["agent"],
                                    decoded_message["statement"]["statement"]
                                   )
        write_value = WriteValueAssertion(instruction=instruction,
                                          variable=variable,
                                          value=value,
                                          id=decoded_message["id"])
        sat, explanation = self.compiler.add_constraint(write_value)
        return sat, decoded_message["id"], explanation


    async def onMessage(self, payload, isBinary):
        assert not isBinary, "Where did the binary message come from" #  turn off in production
        decoded_message = json.loads(payload.decode('utf8'))
        if self.outstanding_work:
            self.compiler =  await self.outstanding_work
            self.outstanding_work = None
        message_type = decoded_message["type"]
        if message_type == "getAllExecutions":
            self.seq_num = int(decoded_message["seq_num"])
            self.executions = self._onGetAllExecutions(decoded_message)
            reply = {"type": "readyListExecutions", "seq_num": self.seq_num}
        elif message_type == "getNextExecution":
            if int(decoded_message["seq_num"]) != self.seq_num:
            #     reply = {"type": "allExecutionsListed"}
                return
            else:
                try:
                    reply = await self.executions.__anext__()
                    reply["seq_num"] = decoded_message["seq_num"]
                except StopAsyncIteration:
                    reply = {"type": "error",
                            "errorMessage": "Requested new execution, but none exists",
                    }
        else:
            try:
                reply = await self.handle[message_type](decoded_message)
            except KeyError as e:
                reply = {
                    "type": "error",
                    "errorMessage": "%s is not a supported message type" % decoded_message["type"]
                }
        self.sendMessage(json.dumps(reply).encode('utf8'))


if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser(description="The m3 server")
    parser.add_argument("--port", "-p", type=int, help="The port, defaults to 5678", default=5678)
    parser.add_argument("--hostname", "-n", type=str, help="The hostname, defaults to 127.0.0.1",
                        default="127.0.0.1")
    parser.add_argument('-v', '--verbose', action='count', default=0)
    args = parser.parse_args()

    host = args.hostname
    port = args.port
    if args.verbose > 0:
        logging.basicConfig(level=logging.INFO)

    factory = WebSocketServerFactory("wss://%s:%d" % (host, port))
    factory.protocol = M3ServerProtocoll
    # Function to accept offers from the client ..
    def accept(offers):
        for offer in offers:
            if isinstance(offer, PerMessageDeflateOffer):
                return PerMessageDeflateOfferAccept(offer)
    factory.setProtocolOptions(perMessageCompressionAccept=accept)

    sc = None
    try:
        sc = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        sc.load_cert_chain('/home/m3-deploy/certificates/cert.pem', '/home/m3-deploy/certificates/privkey.pem')
    except Exception as e:
        logging.warn("couldn't open ssl certs")
        logging.info(str(e))
        sc = None

    loop = asyncio.get_event_loop()
    coro = loop.create_server(factory, host, port, ssl=sc)

    server = loop.run_until_complete(coro)
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass
    finally:
        server.close()
        loop.close()
