"""
Contains functionality to transform TinyPseuCo into z3 formulas
"""

from functools import singledispatch, update_wrapper
from itertools import chain
import typing
from typing import List, Tuple, NamedTuple, Dict, Union, Optional
from enum import Enum, unique, auto, IntFlag
from collections import defaultdict
import logging

from bidict import bidict
import networkx as nx
import z3
from z3 import ForAll, And, Or, Implies, Exists, Not

from pseuco_parser import (BaseExpression, UnaryExpression, BinaryExpression, UnaryOperator,
                           BinaryOperator, IntegerConstant, BooleanConstant, VariableUsage,
                           Statement, EmptyStatement, SimpleStatement, FenceStatement,
                           LockStatement, UnlockStatement, MultiStatement, ITE, VariableAssignment,
                           VariableDeclaration, PrintlnStatement, PseuCoType)

from communication import (Explanation, MemoryModel, InstructionId, OrderAssertion,
                           WriteSeenAssertion, ReadValueAssertion, WriteValueAssertion, Assertion)

@unique
class Reason(Enum):
    """The reason why an order pair was added"""
    ProgramOrder = auto()
    ReadWrite = auto()
    AssertedRead = auto()
    AssertedOrder = auto()

@unique
class CreateNewModel(IntFlag):
    """When generating a new model, specify what gets considered"""
    IgnoreOrder = auto()
    IgnoreWriteSeenRel = auto()

def reason2explanation(reason: Reason):
    if reason is Reason.ProgramOrder:
        return "because of the program order"
    elif reason is Reason.ReadWrite:
        return "because you put them into a read-write relation"
    elif reason is Reason.AssertedRead:
        return "because they are in a read-write relation which justifies your asserted read"
    elif reason is Reason.AssertedOrder:
        return "because you asserted the order"
    else:
        raise NotImplementedError("Unknown enum entry. There shouldn't be any further reason!")


def construct_graph(pairs) -> nx.DiGraph:
    """Turns list of tuple of statements into a graph"""
    g = nx.DiGraph()
    g.add_edges_from(pairs)
    return g


def extract_order_instructions_from_core(core) -> Tuple[str, str, Dict[str, Reason]]:
    def generator():
        for assertion in map(str, core):
            if assertion.startswith("a") and "<" in assertion:
                yield (*assertion.split(" < "), {"reason": Reason.ProgramOrder})
            elif assertion.startswith("manually_asserted_order"):
                orderinfo = assertion.split(":")[1]
                yield (*orderinfo.split("<"), {"reason": Reason.AssertedOrder})
    return generator()

def extract_read_write_pairs_from_core(core, exclude=None) -> Tuple[str, str, Dict[str, Reason]]:
    ret = []
    for assertion in (str(_assertion) for _assertion in core if _assertion != exclude):
        if assertion.startswith("manually_asserted_read_write"):
            splitted =  assertion[len("manually_asserted_read_write"):].split(":")
            read, write = splitted[0], splitted[1]
            ret.append((write, read, {"reason": Reason.ReadWrite}))
    return ret


def methoddispatch(func):
    """ Taken from
    https://stackoverflow.com/questions/24601722/how-can-i-use-functools-singledispatch-with-instance-methods
    Allows single dispatch for methods
    """
    dispatcher = singledispatch(func)

    def wrapper(*args, **kw):  # pylint: disable=missing-docstring
        return dispatcher.dispatch(args[1].__class__)(*args, **kw)
    wrapper.register = dispatcher.register
    update_wrapper(wrapper, func)
    return wrapper


class Z3Compiler:

    def __init__(self, program, timeout=20000) -> None:
        self._context = z3.Context()
        self.solver: z3.Solver = z3.Solver(ctx=self._context)
        self.solver.set(timeout=timeout)
        self.solver.set('core.minimize', True)
        self.statement_enablers: List[z3.z3.ExprRef] = []
        self.global_reads: Dict["Instruction", Tuple["PseuCoVar", z3.IntNumRef]] = {}
        self.global_writes: Dict["Instruction", Tuple["PseuCoVar", z3.IntNumRef]] = {}
        self.fences: List["Instruction"] = []
        self.locks: List[typing.Tuple["Instruction", "PseuCoLock"]] = []
        self.unlocks: List[typing.Tuple["Instruction", "PseuCoLock"]] = []
        self.parent_instructions: List["Instruction"] = []
        self.current_agent_id = 0
        self.agent_id2agent = {}
        self.direct_predecessors = {}
        # defaultdic, so  querying the core is easier and doesn't cause KeyErrors
        default_bool = z3.Bool("dummy_default", ctx=self._context)
        self.named_assertions: Dict[str, "Instruction"] = defaultdict(lambda: default_bool)
        self.memory_model = MemoryModel.SequentialConsistency

        self._local_var_table: typing.Dict[str, z3.ArithRef] = {}
        self._instruction2instruction_id: bidict = bidict()
        self._current_agent_used_id_counter = 0
        self._solver_stack_gauge = 0
        self._manually_added_assertions: List[Assertion] = []
        self._asserted_read_writes: List[z3.BoolRef] = []


        # 1. create the Instruction datatype/enum . It contains one constructor per instruction,
        # that is per statement which does a global read or write. Those consist of every variable
        # declaration in the initialization section, as well as writes and reads in each started
        # agent
        # Furthermore create a datatype for global variables
        # First, Instruction
        Instruction = z3.Datatype("Instruction", ctx=self._context)
        for declaration_counter in range(len(program.initialization.variable_declarations)):
            Instruction.declare("a0_%d" % declaration_counter)
        if not program.initialization.variable_declarations:
            # create a dummy variable as z3 hates empty datatypes
            Instruction.declare("dummyvar")
        startedAgents = program.startedAgentNames.asList()
        for agent_counter, agent_name in enumerate(chain.from_iterable(startedAgents)):
            agent = program.initialization.agent_declarations[agent_name]
            self.agent_id2agent[agent_counter] = agent
            for statement_counter in range(agent.global_operations):
                Instruction.declare("a%d_%d" % (agent_counter+1, statement_counter))
        self.Instruction = Instruction.create()

        # Now, PseuCoVar
        PseuCoVar = z3.Datatype('PseuCoVar', ctx=self._context)
        # every legal PseuCo variable name works as a python identifier, we can make use of that
        # they are also necessarily unique
        for declaration in program.initialization.variable_declarations:
            PseuCoVar.declare(declaration.name)
        if not program.initialization.variable_declarations:
            PseuCoVar.declare("DummyToAvoidCrashes")
        self.PseuCoVar = PseuCoVar.create()

        # now, PseuCoLock
        PseuCoLock = z3.Datatype('PseuCoLock', ctx=self._context)
        for declaration in program.initialization.lock_declarations:
            PseuCoLock.declare(declaration.name)
        if not program.initialization.lock_declarations:
            PseuCoLock.declare("DummyToAvoidCrashes")
        self.PseuCoLock = PseuCoLock.create()

        # 2. Now that Instructions do exist, define relations on them
        self.InstructionOrder = z3.Function("InstructionOrder", self.Instruction, self.Instruction,
                                            z3.BoolSort(ctx=self._context))
        self.ProgramOrder = z3.Function("ProgramOrder", self.Instruction, self.Instruction,
                                            z3.BoolSort(ctx=self._context))
        self.ThreadAffinity = z3.Function("ThreadAffinity", self.Instruction,
                                          z3.IntSort(ctx=self._context))
        # Todo: support more than int for read and write
        self.ReadValue = z3.Function("read", self.Instruction, self.PseuCoVar,
                                     z3.IntSort(ctx=self._context), z3.BoolSort(ctx=self._context))
        self.WriteValue = z3.Function("write", self.Instruction, self.PseuCoVar,
                                      z3.IntSort(ctx=self._context), z3.BoolSort(ctx=self._context))
        # "output" relation, used to get the read-write-relation from the model
        self.ReadWrite = z3.Function("ReadWrite", self.Instruction, self.Instruction)

        # Many properties only have to hold for statements which are actually reachable
        # thus we need to define a predicate for reachability
        self.Reachable = z3.Function("reachable", self.Instruction, z3.BoolSort(ctx=self._context))

        # we need to be able to talk about fences
        self.Fence = z3.Function("Fence", self.Instruction, z3.BoolSort(ctx=self._context))

        # and about Locks
        self.IsLockFor = z3.Function("IsLockFor", self.Instruction, self.PseuCoLock,
                                     z3.BoolSort(ctx=self._context))
        self.IsUnlockFor = z3.Function("IsUnlockFor", self.Instruction, self.PseuCoLock,
                                     z3.BoolSort(ctx=self._context))

        # 3. generate constraints for global variables and their initialization
        for declaration in program.initialization.variable_declarations:
            instr = self._create_and_order_instruction(declaration)
            expr = self.compile_expression(declaration.initial_value)
            self.make_global_write(declaration.name, instr, expr)
        initial_parents = self.parent_instructions

        # 4. Iterate over all started agents, and generate code for them
        for agent_name in chain.from_iterable(startedAgents):
            self.parent_instructions = initial_parents
            self.current_agent_id += 1
            self._current_agent_used_id_counter = 0
            self._local_var_table = {}  # reset local variables
            agent = program.initialization.agent_declarations[agent_name]
            body = agent.functionBody
            for statement in body.statements:
                self.compile_statement(statement)
        self._assert_reads_and_writes()
        self._assert_fences()
        self._assert_locks_and_unlocks()
        self._track_predecessors()
        sat_stat = self.solver.check()
        assert sat_stat == z3.sat, sat_stat


    def _push(self):
        self.solver.push()
        self._solver_stack_gauge += 1

    def _pop(self):
        assert self._solver_stack_gauge > 0
        self.solver.pop()
        self._solver_stack_gauge -= 1

    def _pop_all(self):
        for _ in range(self._solver_stack_gauge):
            self.solver.pop()
        self._solver_stack_gauge = 0

    def _assert_and_track(self, assertion, name):
        b = z3.Bool(name, self._context)
        self.solver.assert_and_track(assertion, b)
        self.named_assertions[name] = b
        return b


    def make_global_read(self, varname: str,
                         instr: "Instruction") -> z3.z3.ArithRef:
        """Takes the variable name and where it is written
        :returns: a variable reperesenting the result of the read
        """
        read_value = z3.Const(str(self.current_agent_id) + str(instr),
                              z3.IntSort(ctx=self._context))
        variable = self.PseuCoVar.__getattribute__(varname)
        self.global_reads[instr] = (variable, read_value)
        return read_value

    def make_global_write(self,
                          varname: str,
                          instr: "Instruction",
                          written_value: z3.z3.ArithRef):
        """Prepares the addition of an global variable read constraint"""
        variable = self.PseuCoVar.__getattribute__(varname)
        self.global_writes[instr] =  (variable, written_value)

    def create_local_pseuco_variable(self, name, var_type: PseuCoType):
        if var_type is PseuCoType.Int:
            new_var = z3.Int(str((self.current_agent_id, name)), ctx=self._context)
        elif var_type is PseuCoType.Boolean:
            new_var = z3.Bool(str((self.current_agent_id, name)), ctx=self._context)
        else:
            raise NotImplementedError("No support for %s" % str(var_type))
        assert name not in self._local_var_table, "no shadowing!"
        self._local_var_table[name] = new_var
        return new_var

    @methoddispatch
    def compile_expression(self, expression: BaseExpression) -> z3.z3.ExprRef:
        raise NotImplementedError("This does not work" + str(type(expression)) + str(expression))

    @compile_expression.register(UnaryExpression)
    def _unary(self, unary_exp: UnaryExpression):
        if unary_exp.operator is UnaryOperator.Plus:
            return self.compile_expression(unary_exp.operand)
        elif unary_exp.operator is UnaryOperator.Minus:
            return -self.compile_expression(unary_exp.operand)
        elif unary_exp.operator is UnaryOperator.Not:
            return not self.compile_expression(unary_exp.operand)
        else:
            raise NotImplementedError("%s is not supported" %
                                      unary_exp.operator)

    @compile_expression.register(IntegerConstant)
    def _intconst(self, c: IntegerConstant) -> z3.z3.IntNumRef:
        return z3.IntVal(c.value, ctx=self._context)

    @compile_expression.register(BooleanConstant)
    def _boolconst(self, c: BooleanConstant) -> z3.z3.BoolRef:
        return z3.BoolVal(c.value, ctx=self._context)

    @compile_expression.register(BinaryExpression)
    def _binexp(self, binary_exp: BinaryExpression):
        lhs = self.compile_expression(binary_exp.lhs)
        rhs = self.compile_expression(binary_exp.rhs)
        op = binary_exp.operator
        if op is BinaryOperator.Plus:
            return lhs + rhs
        elif op is BinaryOperator.Minus:
            return lhs - rhs
        elif op is BinaryOperator.Mult:
            return lhs * rhs
        elif op is BinaryOperator.Div:
            return lhs / rhs
        elif op is BinaryOperator.And:
            return lhs and rhs
        elif op is BinaryOperator.Or:
            return lhs or rhs
        elif op is BinaryOperator.Equal:
            return lhs == rhs
        elif op is BinaryOperator.NEqual:
            return lhs != rhs
        elif op is BinaryOperator.Lt:
            return lhs < rhs
        elif op is BinaryOperator.Lte:
            return lhs <= rhs
        elif op is BinaryOperator.Gt:
            return lhs > rhs
        elif op is BinaryOperator.Gte:
            return lhs >= rhs

    @compile_expression.register(VariableUsage)
    def _varusage(self, var: VariableUsage):
        # NOTE: local_var_table is modified such that globals point to their current value
        return self._local_var_table[var.varname]

    def _create_and_order_instruction(self,
                                      statement: Statement) -> "Instruction":
        instr = self.Instruction.__getattribute__("a%d_%d" % (self.current_agent_id,
                                                              self._current_agent_used_id_counter))
        self._current_agent_used_id_counter += 1
        self._instruction2instruction_id[instr] = InstructionId(self.current_agent_id, statement.id)
        self.solver.add(self.ThreadAffinity(instr) == self.current_agent_id)
        for parent in self.parent_instructions:
            self._assert_and_track(self.InstructionOrder(parent, instr),
                                   f"{parent} < {instr}")
            self._assert_and_track(self.ProgramOrder(parent, instr),
                                   f"{parent} porder {instr}")
        self.parent_instructions = [instr]
        self._assert_and_track(Implies(And(*self.statement_enablers, self._context),
                                       self.Reachable(instr), self._context),
                               f"{self.statement_enablers} ==> {instr}")
        for enabler in self.statement_enablers:
            self._assert_and_track(Implies(self.Reachable(instr), enabler, self._context),
                                   f"{instr} requires {enabler}")
        return instr

    def _handle_globals_in_read(self,
                                instr: "Instruction",
                                expression: BaseExpression):
        assert len(expression.read_global_variables), \
            "After simplification, expression must contain exactly one global variable"
        for varname in expression.read_global_variables:
            val = self.make_global_read(varname, instr)
            self._local_var_table[varname] = val

    @methoddispatch
    def compile_statement(self, statement: Statement):
        raise NotImplementedError("no support for %s" % str(statement))

    @compile_statement.register(EmptyStatement)
    def _empty(self, statement: EmptyStatement):
        return

    @compile_statement.register(FenceStatement)
    def _fence(self, fence: FenceStatement):
        self.fences.append(self._create_and_order_instruction(fence))

    @compile_statement.register(LockStatement)
    def _lock(self, lock: LockStatement):
        self.locks.append((self._create_and_order_instruction(lock),
                           self.PseuCoLock.__getattribute__(lock.lock_name)))

    @compile_statement.register(UnlockStatement)
    def _lock(self, unlock: UnlockStatement):
        self.unlocks.append((self._create_and_order_instruction(unlock),
                             self.PseuCoLock.__getattribute__(unlock.lock_name)))

    @compile_statement.register(SimpleStatement)
    def _simple(self, statement: SimpleStatement):
        assert not statement.expression.read_global_variables, \
            "simple statement must not contain globals after simplification"

    @compile_statement.register(PrintlnStatement)
    def _println(self, println: PrintlnStatement):
        assert not any(expression.read_global_variables for expression in println.expressions), \
            "println statement must not contain global reads after simplification"

    @compile_statement.register(MultiStatement)
    def _multi(self, statement: MultiStatement):
        for statement in statement.statements:
            self.compile_statement(statement)

    @compile_statement.register(ITE)
    def _ite(self, ite: ITE):
        assert not ite.condition.read_global_variables, \
            "Condition must not contain global reads after simplification"
        expr = self.compile_expression(ite.condition)
        self.statement_enablers.append(expr)
        old_parents = self.parent_instructions
        consequence_parents = []
        alternative_parents = []
        self.compile_statement(ite.consequence)
        consequence_parents = self.parent_instructions
        self.parent_instructions = old_parents
        if ite.alternative:
            # flip last condition
            self.statement_enablers[-1] = Not(self.statement_enablers[-1])
            self.compile_statement(ite.alternative)
            alternative_parents = self.parent_instructions
        self.statement_enablers.pop()
        self.parent_instructions = consequence_parents + alternative_parents

    @compile_statement.register(VariableDeclaration)
    def _vardec(self, vardec: VariableDeclaration):
        # WARNING: This only handles local variable declarations
        var = self.create_local_pseuco_variable(vardec.name, vardec.var_type)
        if (vardec.initial_value.read_global_variables):
            instr = self._create_and_order_instruction(vardec)
            self._handle_globals_in_read(instr, vardec.initial_value)
        value = self.compile_expression(vardec.initial_value)
        fact = var == value
        # self.solver.add(fact)
        self._assert_and_track(fact, str(fact))

    @compile_statement.register(VariableAssignment)
    def _varassign(self, assignment: VariableAssignment):
        assert assignment.is_global_write, "Unsupported modification of local"
        # TODO: cannot handle read and write of global in same statement
        assert not assignment.assignment_expression.read_global_variables, \
            "assignment must not contain global reads after simplification"
        written_value = self.compile_expression(
            assignment.assignment_expression)
        instr = self._create_and_order_instruction(assignment)
        self.make_global_write(assignment.var_name, instr, written_value)

    def _track_predecessors(self):
        for instruction in self._instruction2instruction_id:
            predecessor = z3.Const("pre"+str(instruction), self.Instruction)
            pre, notpre = z3.Consts("pre notpre", self.Instruction)
            clause = Implies(
                self.Reachable(instruction),
                Or(
                    ForAll([pre],
                           Implies(
                               self.InstructionOrder(pre, instruction),
                               Or(
                                   pre == instruction,
                                   Not(self.Reachable(pre)),
                                   self._context
                               )
                           )
                           ),
                    Exists([pre],
                           And(
                               self.InstructionOrder(pre, instruction),
                               self.Reachable(pre),
                               pre != instruction,
                               predecessor == pre,
                               Not(Exists([notpre],
                                          And(
                                              self.Reachable(notpre),
                                              notpre != pre,
                                              notpre != instruction,
                                              self.InstructionOrder(notpre, instruction),
                                              self.InstructionOrder(pre, notpre),
                                              self._context)
                                          )
                                   ),
                               self._context
                           ),
                        ),
                    self._context
                )
            )
            self.solver.add(clause)
            self.direct_predecessors[instruction] = predecessor

    def _extract_write_seen_from_model(self, model) -> List[Dict[str, InstructionId]]:
        """Extracts values from ReadWrite"""
        ret = []
        for read  in self.global_reads:
            if z3.is_true(model.evaluate(self.Reachable(read))):
                write = model.evaluate(self.ReadWrite(read))
                ret.append({
                    "write": self._instruction2instruction_id[write],
                    "seen": self._instruction2instruction_id[read]
                })
        return ret

    def _extract_predecessors_from_model(self, model) -> List[Dict[str, InstructionId]]:
        """Converts direct_predecessors from an instruction mapping to a statement mapping"""
        ret = []
        for instr, instr_id in self._instruction2instruction_id.items():
            if instr_id.agent_id == 0 or z3.is_false(model.evaluate(self.Reachable(instr))):
                # initial statements do not have a predecessor
                continue
            pred_instr = model[self.direct_predecessors[instr]]
            pred_id = self._instruction2instruction_id[pred_instr]
            ret.append({
                "statement": instr_id,
                "predecessor": pred_id
            })
        return ret

    def _extract_reachable_from_model(self, model):
        return [instr_id for instr, instr_id in self._instruction2instruction_id.items()
                if z3.is_true(model.evaluate(self.Reachable(instr)))]

    def _assert_reads_and_writes(self):
        """Adds all reads and writes to the solver"""
        a = z3.Const('a', self.Instruction)
        x = z3.Const('x', self.PseuCoVar)
        p = z3.Const('p', z3.IntSort(ctx=self._context))
        # specify read and write function
        reads = ForAll([a, x, p],
                       self.ReadValue(a, x, p) == (
                           Or(*(And(a == b, x == y, p == q, self._context)
                                for (b, (y, q))
                                in self.global_reads.items()), self._context))
                       )
        writes = ForAll([a, x, p],
                        self.WriteValue(a, x, p) == (
                            Or(*(And(a == b, x == y, p == q, self._context)
                                 for (b, (y, q))
                                 in self.global_writes.items()), self._context))
                        )
        self._assert_and_track(reads, "reads")
        self._assert_and_track(writes, "writes")

    def _assert_locks_and_unlocks(self):
        """Specifies lock and unlock statements"""
        a = z3.Const('a', self.Instruction)
        l = z3.Const('l', self.PseuCoLock)
        locks = ForAll([a, l],
                       self.IsLockFor(a, l) == (
                           Or(*(And(a == b, l == ll, self._context)
                                for (b, ll) in self.locks),
                              self._context)
                       ))
        unlocks = ForAll([a, l],
                         self.IsUnlockFor(a, l) == (
                             Or(*(And(a == b, l == ll, self._context)
                                  for (b, ll) in self.unlocks),
                                self._context)
                         ))
        self._assert_and_track(locks, "locks")
        self._assert_and_track(unlocks, "unlocks")

    def _assert_fences(self):
        a = z3.Const('a', self.Instruction)
        fences = ForAll([a],
                        self.Fence(a) ==
                            Or(*(a == b for b in self.fences), self._context)
                        )
        self._assert_and_track(fences, "fences")

    def _has_obvious_ootr(self, model) -> Tuple[bool, Optional[str]]:
        self.solver.push()
        current_ws = [self.ReadWrite(instr) == model.evaluate(self.ReadWrite(instr))
                      for instr, (_, value) in self.global_reads.items()
                      if z3.is_true(model.evaluate(self.Reachable(instr)))
                      ]
        current_reads = (value == model[value].as_long()
                         for instr, (_, value) in self.global_reads.items()
                         if z3.is_true(model.evaluate(self.Reachable(instr))))
        self.solver.add(And(*current_ws, self._context))
        self.solver.add(Not(And(*current_reads, self._context)))
        has_ootr = self.solver.check() == z3.sat
        self.solver.pop()
        if has_ootr:
            return has_ootr, current_ws
        else:
            return has_ootr, None

    def _find_new_model(self, model, counter, strategy):
        # if the new model can provide a new set of read values, we are always interested in it
        current_reads = (value == model[value].as_long()
                         for instr, (_, value) in self.global_reads.items()
                         if z3.is_true(model.evaluate(self.Reachable(instr))))
        current_read_write = []
        # if we care about the ReadValue relationship, also add models where it differs (the
        # default)
        if not (strategy & CreateNewModel.IgnoreWriteSeenRel):
            current_read_write = (self.ReadWrite(instr) == model.evaluate(self.ReadWrite(instr))
                        for instr, (_, value) in self.global_reads.items()
                        if z3.is_true(model.evaluate(self.Reachable(instr)))
                                )
        # if we care about the different orders (TODO: not implemented):
        current_order = []
        if not(strategy & CreateNewModel.IgnoreOrder):
            pre, notpre = z3.Consts("pre notpre", self.Instruction)
            current_order = [
                self.InstructionOrder(model[self.direct_predecessors[instr]], instr)
                for instr, instr_id in self._instruction2instruction_id.items()
                if not (instr_id.agent_id == 0 or z3.is_false(model.evaluate(self.Reachable(instr))))
            ]
        self._assert_and_track(Not(And(*current_reads, *current_read_write, *current_order, self._context)),
                               "counter %s" % counter,)
        return self.solver.check()


    def _add_incosistent_axioms(self):
        """Create the axioms for inconsistency (any written value can be read if var matches)"""
        # NOTE: Impossible to avoid out-of-thin-air-reads
        logging.info("adding inconsistency axioms")
        a, b, c = z3.Consts('a b c', self.Instruction)
        x, y = z3.Consts('x y', self.PseuCoVar)
        p, q = z3.Consts('p q', z3.IntSort(ctx=self._context))

        incosist_axioms = {
            "sq_write-seen": ForAll([a, x, p], Implies(
                And(self.ReadValue(a, x, p), self.Reachable(a), self._context),
                Exists([b], And(self.WriteValue(b, x, p),
                                self.Reachable(b),
                                self.ReadWrite(a) == b,
                                self._context
                                ))
            ))
        }
        for name, fact in incosist_axioms.items():
            self._assert_and_track(fact, name)


    def _add_seqconsist_axioms(self):
        """Creates the axioms of sequential consistency"""
        logging.info("adding sequential consistency axioms")
        a, b, c = z3.Consts('a b c', self.Instruction)
        x, y = z3.Consts('x y', self.PseuCoVar)
        p, q = z3.Consts('p q', z3.IntSort(ctx=self._context))

        seqconsist_axioms = {
            "sq_total": ForAll([a, b], Implies(
                And(self.Reachable(a), self.Reachable(b), self._context),
                Or(self.InstructionOrder(a, b), self.InstructionOrder(b, a), self._context)
            )),  # order is total on reachable statements
            "sq_transitive": ForAll([a, b, c], Implies(
                And(self.InstructionOrder(a, b), self.InstructionOrder(b, c), self._context),
                self.InstructionOrder(a, c)
            )),  # transitive
            "sq_antisymmetric": ForAll([a, b], Implies(
                And(self.InstructionOrder(a, b), self.InstructionOrder(b, a), self._context),
                a == b
            )),
            "sq_write-seen": ForAll([a, x, p], Implies(
                And(self.ReadValue(a, x, p), self.Reachable(a), self._context),
                Exists([b], And(self.WriteValue(b, x, p),
                                self.Reachable(b),
                                self.ReadWrite(a) == b,
                                self.InstructionOrder(b, a),
                                Not(
                                    Exists([c, q], And(
                                        self.Reachable(c),
                                        c != b,
                                        self.InstructionOrder(b, c),
                                        self.InstructionOrder(c, a),
                                        self.WriteValue(c, x, q),
                                        self._context
                                    ))
                                ),
                                self._context
                                ))
            )),
        }
        for name, fact in seqconsist_axioms.items():
            self._assert_and_track(fact, name)

    def _add_weak_axioms(self, use_fence=False, use_lock=False):
        """Creates the axioms of sequential consistency"""
        logging.info("adding weak consistency axioms")
        a, b, c, d, e = z3.Consts('a b c d e', self.Instruction)
        x, y = z3.Consts('x y', self.PseuCoVar)
        p, q = z3.Consts('p q', z3.IntSort(ctx=self._context))
        l = z3.Const('l', self.PseuCoLock)


        weak_order_axioms = {
            "w_reflexive": ForAll([a] , Implies(self.Reachable(a),
                                self.InstructionOrder(a, a))),  # order is reflexive
            "w_transitive": ForAll([a, b, c], Implies(
                And(self.InstructionOrder(a, b), self.InstructionOrder(b, c), self._context),
                self.InstructionOrder(a, c)
            )),  # transitive
            "w_antisymmetric": ForAll([a, b], Implies(
                And(self.InstructionOrder(a, b), self.InstructionOrder(b, a), self._context),
                a == b
            )),
            "w_write-seen": ForAll([a, x, p], Implies(
                And(self.ReadValue(a, x, p), self.Reachable(a), self._context),
                Exists([b], And(self.WriteValue(b, x, p),
                                self.Reachable(b),
                                self.ReadWrite(a) == b,
                                Not(self.InstructionOrder(a, b)),
                                Not(
                                    Exists([c, q], And(
                                        self.Reachable(c),
                                        c != b,
                                        self.InstructionOrder(b, c),
                                        self.InstructionOrder(c, a),
                                        self.WriteValue(c, x, q),
                                        self._context
                                    ))
                                ),
                                self._context
                                ))
            )),
        }

        for name, fact in weak_order_axioms.items():
            self._assert_and_track(fact, name)

        if use_fence:
            fence_axioms = {
                "w_fence-order": ForAll([a, b, c], Implies(
                    And(self.Fence(c), self.InstructionOrder(c, b), self.ReadWrite(a) == b, self._context),
                    self.InstructionOrder(c, a),
                    self._context
                )),
            }

            for name, fact in fence_axioms.items():
                self._assert_and_track(fact, name)

        if not use_lock:
            return

        lock_axioms = {
            "p_reflexive": ForAll([a], Implies(self.Reachable(a),
                                               self.ProgramOrder(a, a))),  # order is reflexive
            "p_transitive": ForAll([a, b, c], Implies(
                And(self.ProgramOrder(a, b), self.ProgramOrder(b, c), self._context),
                self.ProgramOrder(a, c)
            )),  # transitive
            "p_antisymmetric": ForAll([a, b], Implies(
                And(self.ProgramOrder(a, b), self.ProgramOrder(b, a), self._context),
                a == b
            )),
            "w_lock-order": ForAll([l, a, b, c, d], Implies(
                And(self.IsLockFor(a, l), self.IsUnlockFor(b, l),
                    self.IsLockFor(c, l), self.IsUnlockFor(d, l),
                    self.Reachable(a), self.Reachable(b), self.Reachable(c), self.Reachable(d),
                    a != c, b != d, self.ProgramOrder(a, b), self.ProgramOrder(c, d),
                    self.ThreadAffinity(a) == self.ThreadAffinity(b),
                    self.ThreadAffinity(c) == self.ThreadAffinity(d),
                    self.ThreadAffinity(a) != self.ThreadAffinity(c),
                    self.ThreadAffinity(b) != self.ThreadAffinity(d),
                    Not(
                        Exists([e], And(
                            self.Reachable(e),
                            Or(self.IsLockFor(e, l), self.IsUnlockFor(e, l), self._context),
                            e != a, e != b,
                            self.ThreadAffinity(e) == self.ThreadAffinity(a),
                            self.ProgramOrder(a, e),
                            self.ProgramOrder(e, b),
                            self._context
                        ))
                    ),
                    Not(
                        Exists([e], And(
                            self.Reachable(e),
                            Or(self.IsLockFor(e, l), self.IsUnlockFor(e, l), self._context),
                            e != c, e != d,
                            self.ThreadAffinity(e) == self.ThreadAffinity(c),
                            self.ProgramOrder(c, e),
                            self.ProgramOrder(e, d),
                            self._context
                        ))
                    ),
                    self._context
                   ),
                Or(self.InstructionOrder(b, c), self.InstructionOrder(d, a), self._context),
                self._context
            ))
        }
        for name, fact in lock_axioms.items():
            self._assert_and_track(fact, name)



    def list_executions(self, strategy):
        logging.info("listing all executions")
        logging.debug(self.solver)
        is_sat = self.solver.check()
        if is_sat == z3.unknown:
            if self.solver.reason_unknown() == 'timeout':
                logging.warn(self.solver.statistics())
                raise ResourceWarning("Timeout occured")
            logging.warn("Cannot decide given program. WTH did you do?")
        assert  is_sat != z3.unsat, "Program without constraints should always be sat or unknown"
        counter = 0
        self._push()
        while is_sat == z3.sat and counter < 50:
            logging.info(f"{counter}")
            counter += 1
            model = self.solver.model()
            execution = []
            has_ootr = False
            if self.memory_model != MemoryModel.SequentialConsistency:
                has_ootr, bad_rw = self._has_obvious_ootr(model)
                if has_ootr:
                    self.solver.add(Not(And(*bad_rw, self._context)))
            for (instr, (var, value)) in self.global_reads.items():
                read_instr_id = self._instruction2instruction_id[instr]
                if z3.is_false(model.evaluate(self.Reachable(instr))):
                    continue  # skip unreachable reads
                read_value = model[value].as_long()
                execution.append({
                    "agent_statement_id": read_instr_id,
                    "read_value": read_value,
                    "read_var": str(var),
                })
            yield {
                "counter": counter,
                "reads": execution,
                "ordering": self._extract_predecessors_from_model(model),
                "reachable": self._extract_reachable_from_model(model),
                "writeSeen": self._extract_write_seen_from_model(model),
                "has_ootr": has_ootr,
            }
            is_sat = self._find_new_model(model, counter, strategy)
        self._pop()
        if counter == 50:
            logging.warning("too many reads!?!")

    def _replay_manually_added_assertions(self) -> Tuple[bool, int]:
        logging.info("replaying assertions")
        self._manually_added_assertions, assertions = [], self._manually_added_assertions
        acked = []
        nacked = None
        for assertion in assertions:
            if nacked:
                # preserve existing asertions, even if we already encountered a nack
                self._manually_added_assertions.append(assertion)
                continue
            sat, explanation = self.add_constraint(assertion)
            if not sat:
                nacked = {"id": assertion.id, "explanation": explanation}
            else:
                acked.append(assertion.id)
        return acked, nacked


    def switch_manually_added_assertion(self, oldpos, newpos) -> Tuple[bool, int]:
        self._manually_added_assertions[oldpos], self._manually_added_assertions[newpos] = (
            self._manually_added_assertions[newpos], self._manually_added_assertions[oldpos]
        )
        self._pop_all()
        self._push()
        if self.memory_model is MemoryModel.SequentialConsistency:
            self._add_seqconsist_axioms()
        elif self.memory_model is MemoryModel.WeakConsistency:
            self._add_weak_axioms()
        elif self.memory_model is MemoryModel.WeakConsistencyFences:
            self._add_weak_axioms(use_fence=True)
        elif self.memory_model is MemoryModel.WeakConsistencyLocks:
            self._add_weak_axioms(use_lock=True)
        return self._replay_manually_added_assertions()


    def _is_there_always_a_write_between_awrite_and_read(self, read, awrite, variable):
        self._push()
        anyval = z3.Int("anyval", self._context)
        write = z3.Const("write", self.Instruction)
        nothing_in_between = z3.Not(Exists([write], And(
            write != awrite,
            self.Reachable(write),
            self.ReadWrite(read) == write,
            self.ProgramOrder(awrite, write),
            self._context
        )))
        self.solver.add(nothing_in_between)
        valid = (self.solver.check() == z3.unsat)
        self._pop()
        return valid

    def _readers2matching_writers(self, core):
        self._push()
        read_instrs = []
        all_read_write_pairs = []
        for stringified_assertion in map(str, core):
            if stringified_assertion.startswith("manually_asserted_read:ReadValue"):
                instr_str = stringified_assertion[len("manually_asserted_read:ReadValue("):].split(",")[0]
                instr = self.Instruction.__getattribute__(instr_str)
                read_instrs.append(instr)
        sat = self.solver.check()
        assert(sat == z3.sat)
        while sat == z3.sat:
            model = self.solver.model()
            read_write_pairs = [(reader, model.evaluate(self.ReadWrite(reader)))
                                for reader in read_instrs]
            all_read_write_pairs.append(read_write_pairs)
            current_model = (self.ReadWrite(reader)==writer for (reader,writer) in read_write_pairs)
            self.solver.add(Not(And(*current_model, self._context)))
            sat = self.solver.check()
        self._pop()
        return all_read_write_pairs


    def set_memory_model(self, memory_model: MemoryModel) -> Tuple[bool, int]:
        self.memory_model = memory_model
        self._pop_all()
        self._push()
        if memory_model is MemoryModel.SequentialConsistency:
            self._add_seqconsist_axioms()
        elif memory_model is MemoryModel.WeakConsistency:
            self._add_weak_axioms()
        elif memory_model is MemoryModel.WeakConsistencyFences:
            self._add_weak_axioms(use_fence=True)
        elif memory_model is MemoryModel.WeakConsistencyLocks:
            self._add_weak_axioms(use_lock=True)
        else:
            raise KeyError("unkown memory model: {memory_model}")
        return self._replay_manually_added_assertions()

    def remove_assertion(self, assertion_id: int):
        self._manually_added_assertions = [a for a in self._manually_added_assertions
                                           if a.id != assertion_id]
        # calling set_memory_model is not the most elegant thing to do here, but it works
        # as  it removes all assertions from the solver and then readds those that we currently have
        self.set_memory_model(self.memory_model)



    def generate_explanation_from_graph(self, g: nx.DiGraph, source, target):
        substitutions = set()
        try:
            cycle = nx.find_cycle(g)
            explanation = "You've created a cycle: <br/>"
            for (source, target) in cycle[:-1]:
                substitutions.add(source)
                substitutions.add(target)
                reason = reason2explanation(g[source][target]['reason'])
                explanation += f"{source} happens before {target} {reason}<br/>"
                source, target = cycle[-1]
            reason = reason2explanation(g[source][target]['reason'])
            explanation += f"But {source} happens before {target} {reason}"
            return Explanation(explanation, "", [s[1:].split("_") for s in  substitutions])
        except nx.exception.NetworkXNoCycle:
            print(20*"===")
            print(g.nodes())
            print(source, target)
            print(20*"===")
            try:
                srcs = [src for (src, degree) in g.in_degree(g.nodes()) if degree == 0]
                trgts = [trgt for (trgt, degree) in g.out_degree(g.nodes()) if degree == 0]
                assert(len(srcs) == 1)
                assert(len(trgts) == 1)
                if srcs[0] != source:
                    source = srcs[0]
                    print("Warning: Changed source")
                if trgts[0] != target:
                    target = trgts[0]
                    print("Warning: Changed target")
                print(source, target)
                paths = nx.all_simple_paths(g, source, target)
            except Exception as e:
                print(e)
                return "OOPs"

            explanation = ""
            for path in paths:
                if path == [source, target]:
                    continue
                src = None
                trgt = path[0]
                latest_write = None
                def getvar(instr):
                    try:
                        return self.global_writes[self.Instruction.__getattribute__(instr)][0]
                    except KeyError:
                        return None
                var  = getvar(source)
                for node in path[1:]:
                    src = trgt
                    trgt = node
                    reason = reason2explanation(g[src][trgt]['reason'])
                    if getvar(src) == var:
                        latest_write = src
                    explanation += f"{src} happens before {trgt} {reason}<br/>"
                    substitutions.add(src)
                    substitutions.add(trgt)
                explanation += f"You asserted that {source} and {target} should be in a read-write relation.<br/>"
                explanation += "This would however be a read from the non-immediate past<br/>"
                explanation += f"As {latest_write} is a more recent write"
                substitutions.add(source)
                substitutions.add(target)
                substitutions.add(latest_write)
            return Explanation(explanation, "", [s[1:].split("_") for s in  substitutions])

    def _add_WriteSeenAssertion(self, constraint: WriteSeenAssertion) -> Tuple[bool,
                                                                               Optional[Explanation]]:
        explanation: Optional[Explanation] = None
        challenge = ""
        variable = constraint.variable
        write = self._instruction2instruction_id.inv[constraint.writer]
        read = self._instruction2instruction_id.inv[constraint.reader]
        write_is_reachable = self._assert_and_track(self.Reachable(write),
                                        f"manually_asserted_reachable_write:{write}:{self._solver_stack_gauge}")
        read_is_reachable = self._assert_and_track(self.Reachable(read),
                                        f"manually_asserted_reachable_read:{read}:{self._solver_stack_gauge}")
        anyval = z3.Int("anyval", self._context)
        var = self.PseuCoVar.__getattribute__(variable)
        rw = Exists([anyval], And(
            self.ReadValue(read, var, anyval),
            self.WriteValue(write, var, anyval),
            self.ReadWrite(read) == write,
            self._context
        ))
        read_write_is_valid =  self._assert_and_track(rw,
                                f"manually_asserted_read_write{read}:{write}:{var}:{self._solver_stack_gauge}")
        sat = self.solver.check() == z3.sat
        if sat:
            # early return
            return sat, Explanation
        # if unsat, generate explanation
        core = self.solver.unsat_core()
        self._pop()
        print(core)
        pass
        # 1. reader is unreachable
        if read_is_reachable in core and any(("requires " in x for x in map(str, core))):
            explanation = "Asserted read is not reachable"
            for assertion in map(str, core):
                if "requires" in assertion:
                    _, condition = assertion.split("requires ")
                    print(condition)
                    if condition.startswith("Not"):
                        condition = condition.strip("Not(").rstrip(")")
                        explanation += f"as it requires {condition} to be false, but it is true"
                        challenge = f"Refine the execution so that {condition} is false"
                    else:
                        explanation += f"as it requires {condition} to be true, but it is false"
                        challenge = f"Refine the execution so that {condition} is true"
                    break
            explanation = Explanation(explanation, challenge)
        # 2. writer is unreachable
        elif write_is_reachable in core:
            explanation_str = "Asserted write is not reachable"
            for assertion in map(str, core):
                if "requires" in assertion:
                    _, condition = assertion.split("requires ")
                    print(condition)
                    if condition.startswith("Not"):
                        condition = condition.strip("Not(").rstrip(")")
                        explanation_str += f"as it requires {condition} to be false, but it is true"
                        challenge = f"Refine the execution so that {condition} is false"
                    else:
                        explanation_str += f"as it requires {condition} to be true, but it is false"
                        challenge = f"Refine the execution so that {condition} is true"
                    break
            explanation = Explanation(explanation, challenge)
        # 3. sequential constistent, and there is an order cycle
        #    - check for sq_antisymmetric in core?
        inbetween = self._is_there_always_a_write_between_awrite_and_read(read, write, variable)
        print("always in between: ", inbetween)
        if ( explanation is None and (self.named_assertions["sq_write-seen"] in core or
                self.named_assertions["w_fence-order"] in core) and read_write_is_valid in core and
            (self.named_assertions["sq_transitive"] in core or
                self.named_assertions["sq_antisymmetric"] in core)
            ):
            # convert core into list
            core_list = list(core)
            # use _extract_order_instructions_from_core
            ordering = list(extract_order_instructions_from_core(core_list))
            # add imposed order from WriteSeenAssertions
            ordering += extract_read_write_pairs_from_core(core_list)
            # Plan: Because we are inside the write-seen assertion error handler, we know that
            # all previously asserted read values were possible
            # As we undid the last faulty assertion, the solver is now in a state where we can
            # query it which justifying writes would be valid
            reader_writer_pairs = self._readers2matching_writers(core)
            if reader_writer_pairs:  # TODO: handle ALL possibilities
                ordering += map(lambda x: (str(x[1]), str(x[0]), {"reason": Reason.AssertedRead}), reader_writer_pairs[0])
            g = construct_graph(ordering)
            explanation = self.generate_explanation_from_graph(g, str(write), str(read))
        # 4. read from non-immediate past
        elif (explanation is None and self.named_assertions["sq_write-seen"] in core or
            self.named_assertions["w_write-seen"] in core) and read_write_is_valid in core:
            explanation = (f"{write} happens after {read} in the program order, which violates"
            " the constraint for the write-seen relationship that you should ")
            if self.memory_model is MemoryModel.SequentialConsistency:
                explanation += "should only read from past"
            else:
                explanation += "should not read from the future"
            explanation = Explanation(explanation, "", [write, read])
        #    pop to sat state, and ask solver if
        #    ∃w. constraint.writer < w < constraint.reader is valid, that is, if
        #    Not(∃w. constraint.writer < w < constraint.reader) is sat
        # 5. read wrong value
        elif (explanation is None and self.named_assertions["writes"] in core and self.named_assertions["reads"] in core
                and any((s.startswith("manually_asserted_read:") for s in map(str, core)))):
            for s in map(str, core):
                if s.startswith("manually_asserted_read:"):
                    _, read_value, *_ = s.split(":")
                    number = read_value.split(",")[-1].rstrip(")")
                    break
            explanation = Explanation(f"You previously asserted that {number} should be read. "\
                            "However, the write you selected for this write-seen relationship "
                            "does not provide this value.") # TODO: which values does it contain
        # 6. there was already another write-seen relation with the same variable
        elif (explanation is None and sum(1 for _ in (s.startswith(f"manually_asserted_read_write{read}:") for
                                        s in map(str, core))) >= 2):
            explanation_message = "You have already addded a conflicting write-seen relationship!"
            explanation = Explanation(explanation_message)
        else:
            explanation = Explanation()
        return sat, explanation

    def _add_ReadValueAssertion(self, constraint: WriteSeenAssertion) -> Tuple[bool,
                                                                               Optional[Explanation]]:
        explanation = None
        instr = self._instruction2instruction_id.inv[constraint.instruction]
        var = self.PseuCoVar.__getattribute__(constraint.variable)
        value = z3.IntVal(constraint.value, self._context)
        self._assert_and_track(self.Reachable(instr),
                                        f"manually_asserted_reachable:{instr}:{self._solver_stack_gauge}")
        self._assert_and_track(self.ReadValue(instr, var, value),
                                        f"manually_asserted_read:ReadValue({instr},{var},{value}):{self._solver_stack_gauge}")
        sat = self.solver.check() == z3.sat
        if not sat:
            core = self.solver.unsat_core()
            self._pop()
            print(core)
            # consider all cases
            # 1. read nonsensical value which is written nowhere
            if (len(core) == 4 and
                    self.named_assertions["writes"] in core and
                    (self.named_assertions["w_write-seen"] in core or self.named_assertions["sq_write-seen"] in core)):
                explanation = Explanation(f"You asserted that {constraint.variable} should be "
                                        f"{constraint.value}, but there is no write of "
                                        f"{constraint.value}. If you still think that this is a "
                                        "legal value, please provide a write-seen relation to "
                                        "justify it", f"Provide a write-seen relation to justify a"
                                          f" write of {constraint.value}")
            # 2. read value which could be written, but:
            else:
                explanation = Explanation("There does not exist a write which would justify the "
                                          f"write of {constraint.value}. If you think otherwise, "
                                          "please provide a suitable write-seen relationship to "
                                          "justify it", f"Provide a write-seen relation to justify a"
                                          f" write of {constraint.value}")
            # TODO: provide hits for cases below
            # 2a) is unreachable and not actually written
            # 2b) violates the order
        return sat, explanation

    def _add_WriteValueAssertion(self, constraint: WriteValueAssertion) -> Tuple[bool,
                                                                               Optional[Explanation]]:
        explanation: Optional[Explanation]  = None
        instr = self._instruction2instruction_id.inv[constraint.instruction]
        var = self.PseuCoVar.__getattribute__(constraint.variable)
        value = z3.IntVal(constraint.value, self._context)
        is_reachable = self._assert_and_track(self.Reachable(instr),
                                              f"manually_asserted_reachable:{instr}:{self._solver_stack_gauge}")
        written_value = self._assert_and_track(self.WriteValue(instr, var, value),
                                               f"manually_asserted_write:WriteValue({instr},{var},{value}):{self._solver_stack_gauge}")
        sat = self.solver.check() == z3.sat
        if sat:
            return sat, explanation
        core = self.solver.unsat_core()
        self._pop()
        print(core)
        # consider all cases:
        # 1. wrote constant, asserted something else
        if (self.named_assertions["writes"] in core and written_value in core):
            explanation = Explanation("Written value does not match constant")
        # 2. not a constant, but asserted value can never occur
        # 3. wrote value which could occur, if there weren't previous assertions
        return sat, explanation

    def _add_OrderAssertion(self, constraint: OrderAssertion) -> Tuple[bool,
                                                                       Optional[Explanation]]:
        explanation: Optional[Explanation] = None
        prev_instr = self._instruction2instruction_id.inv[constraint.predecessor]
        next_instr = self._instruction2instruction_id.inv[constraint.successor]
        prev_is_reachable = self._assert_and_track(self.Reachable(prev_instr),
                                        f"manually_asserted_reachable:{prev_instr}:{self._solver_stack_gauge}0")
        next_is_reachable = self._assert_and_track(self.Reachable(next_instr),
                                        f"manually_asserted_reachable:{next_instr}:{self._solver_stack_gauge}1")
        self._assert_and_track(self.InstructionOrder(prev_instr, next_instr) ,
                               f"manually_asserted_order:{prev_instr}<{next_instr}")
        sat = self.solver.check() == z3.sat
        if sat:
            return sat, explanation
        core = self.solver.unsat_core()
        self._pop()
        print(core)
        challenge = ""
        explanation = "M3 cannot explain why this is impossible :-("
        # 1. prev is unreachable
        if prev_is_reachable in core and any(("requires " in x for x in map(str, core))):
            explanation = "The first operation is unreachable "
            for assertion in map(str, core):
                if "requires" in assertion:
                    _, condition = assertion.split("requires ")
                    print(condition)
                    if condition.startswith("Not"):
                        condition = condition.strip("Not(").rstrip(")")
                        explanation += f"as it requires {condition} to be false, but it is true"
                        challenge = f"Refine the execution so that {condition} is false"
                    else:
                        explanation += f"as it requires {condition} to be true, but it is false"
                        challenge = f"Refine the execution so that {condition} is true"
                    break
        # 2. writer is unreachable
        elif next_is_reachable in core:
            explanation = "The second operation is not reachable"
            for assertion in map(str, core):
                if "requires" in assertion:
                    _, condition = assertion.split("requires ")
                    print(condition)
                    if condition.startswith("Not"):
                        condition = condition.strip("Not(").rstrip(")")
                        explanation += f"as it requires {condition} to be false, but it is true"
                        challenge = f"Refine the execution so that {condition} is false"
                    else:
                        explanation += f"as it requires {condition} to be true, but it is false"
                        challenge = f"Refine the execution so that {condition} is true"
                    break
        # check for cycle?
        elif (self.named_assertions["sq_write_seen"] in core and
              (self.named_assertions["sq_transitive"] in core or
               self.named_assertions["sq_antisymmetric"] in core)):
            # TODO
            # convert core into list
            core_list = list(core)
            # use _extract_order_instructions_from_core
            ordering = list(extract_order_instructions_from_core(core_list))
            # add imposed order from WriteSeenAssertions
            ordering += extract_read_write_pairs_from_core(core_list)
            # Plan: Because we are inside the write-seen assertion error handler, we know that
            # all previously asserted read values were possible
            # As we undid the last faulty assertion, the solver is now in a state where we can
            # query it which justifying writes would be valid
            reader_writer_pairs = self._readers2matching_writers(core)
            if reader_writer_pairs:  # TODO: handle ALL possibilities
                ordering += map(lambda x: (str(x[1]), str(x[0]), {"reason": Reason.AssertedRead}), reader_writer_pairs[0])
            g = construct_graph(ordering)
            explanation = self.generate_explanation_from_graph(g, str(write), str(read))
        return sat, Explanation(explanation, challenge)


    def add_constraint(self, constraint: Assertion) -> Tuple[bool, Optional[Explanation]]:
        """addds an constraint if satisfiable and returns it sat status"""
        explanation = None
        sat = None
        self._push()
        if isinstance(constraint, WriteSeenAssertion):
            sat, explanation =  self._add_WriteSeenAssertion(constraint)
        elif isinstance(constraint, ReadValueAssertion):
            sat, explanation = self._add_ReadValueAssertion(constraint)
        elif isinstance(constraint, WriteValueAssertion):
            sat, explanation = self._add_WriteValueAssertion(constraint)
        elif isinstance(constraint, OrderAssertion):
            sat, explanation = self._add_OrderAssertion(constraint)

        if not sat:
            print(explanation)
        self._manually_added_assertions.append(constraint)
        return sat, explanation
